/*
** Module   : ContainerTree.java
** Abstract : The tree Widget Browser component.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
** 002 HC  20141118 Added object identity hash-code property.
** 003 HC  20141130 Added the capability to display widget config fields.
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.event.ActionListener;
import java.lang.reflect.*;
import java.util.*;

import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.*;

import org.apache.commons.lang.builder.*;

import com.goldencode.p2j.ui.WidgetConfig;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The tree Widget Browser component.
 */
@SuppressWarnings("serial")
public class ContainerTree 
extends JTree 
{	
	/**
	 * Full class names flag.
	 */
   private boolean showFullClassName = false;

	/**
	 * Event listener.
	 */
   private ActionListener actionListener;
	
	/**
	 * Related property table components.
	 */
   private PropertyTable table;
	
	/**
	 * The root container.
	 */
   private Container<?> container;

	/**
	 * The constructor.
	 * 
	 * @param   container
	 *          The root container.
	 * @param   table
	 *          The related property table component.
	 */
   public ContainerTree(Container<?> container, final PropertyTable table) 
	{
		super();
		this.table = table;
		this.container = container;
		
		setModel(new DefaultTreeModel(sortNodes(buildTreeModel(container))));

		// disable multiple selection
		getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

		addTreeSelectionListener(new TreeSelectionListener() 
		{
			public void valueChanged(TreeSelectionEvent e) 
			{
				DefaultMutableTreeNode node = (DefaultMutableTreeNode) getLastSelectedPathComponent();

				if (node == null || !(node.getUserObject() instanceof WidgetWrapper)) 
				{
					return;
				}

				table.setTableContent(((WidgetWrapper) node.getUserObject()).getProperties());
				
				if (actionListener != null)
				{
				   actionListener.actionPerformed(
				         new Event(this, 
				                   Event.WIDGET_SELECTION_CHANGED, 
				                   ((WidgetWrapper) node.getUserObject()).getWidget()));
				}
			}
		});
	}
	
	/**
	 * Sets the event listener.
    *
	 * @param   actionListener
	 *          Valid {@link ActionListener} reference.
	 */
   public void setActionListener(ActionListener actionListener)
   {
      this.actionListener = actionListener;
   }

   /**
    * Returns the state of show full class names flag.
    * 
    * @return  The show full class name flag.
    */
   public boolean isShowFullClassName()
	{
		return showFullClassName;
	}

   /**
    * Sets the state of show full class names flag.
    * 
    * @param   showFullClassName
    *          The show full class name flag.
    */
	public void setShowFullClassName(boolean showFullClassName) 
	{
		this.showFullClassName = showFullClassName;
	}

	/**
    * Refreshes (rebuilds) the tree nodes.
    */
   public void refresh()
   {
      // store selected node
      DefaultMutableTreeNode selectedNode = (DefaultMutableTreeNode) getLastSelectedPathComponent();
   
      // store expanded widgets
      List<Widget<?>> expandedWidgets = getExpandedWidgets();
   
      // rebuild tree
      ((DefaultTreeModel) getModel()).setRoot(sortNodes(buildTreeModel(container)));
   
      // restore expanded nodes
      expandRows(findNodes(expandedWidgets));
   
      // restore selected node
      TreePath result = null;
      if (selectedNode instanceof DefaultMutableTreeNode
            && ((DefaultMutableTreeNode) selectedNode).getUserObject() instanceof WidgetWrapper)
      {
         result = findNode(((WidgetWrapper) ((DefaultMutableTreeNode) selectedNode).getUserObject())
                     .getWidget());
      }
   
      if (result == null)
      {
         // clear table data - selected node doesn't exist
         table.resetProperties();
         return;
      }
   
      setSelectionPath(result);
   }

   /**
    * Collapse all tree nodes.
    */
   public void collapseAll()
   {
      for (int i = getRowCount() - 1; i >= 1; i--) 
      {
         collapseRow(i);
      }
   }

   /** 
    * Expand all tree nodes.
    */
   public void expandAll()
   {
      for (int i = 0; i < getRowCount(); i++) 
      {
         expandRow(i);
      }   
   }

   /**
    * Returns the related tree path for the supplied widget.
    * 
    * @param   subject
    *          The {@link Widget} reference.
    *          
    * @return  {@link TreePath} instance or <code>null</code> if not found.
    */
   public TreePath findNode(Widget<?> subject)
   {
      List<Widget<?>> subjects = new ArrayList<>();
      subjects.add(subject);
      List<TreePath> result = findNodes(subjects);
      return result == null || result.isEmpty() ? null : result.get(0);
   }

   /**
    * Checks whether the supplied node is selected.
    * 
    * @param   widget
    *          The {@link Widget} reference.
    *          
    * @return  <code>true</code> if selected, <code>false</code> otherwise.
    */
   public boolean isSelected(Widget<?> widget)
   {
      DefaultMutableTreeNode selectedNode = 
            (DefaultMutableTreeNode) getLastSelectedPathComponent();
      if (selectedNode == null || !(selectedNode.getUserObject() instanceof WidgetWrapper))
      {
         return false;
      }
      
      return Objects.equals(widget, ((WidgetWrapper) selectedNode.getUserObject()).getWidget());
   }

   /**
	 * The worker method responsible for building the tree nodes in the tree component.
	 * 
	 * @param   container
	 *          The root container.
	 * 
	 * @return  The root tree node.
	 */
	private DefaultMutableTreeNode buildTreeModel(Container<?> container) 
	{
		DefaultMutableTreeNode root = new DefaultMutableTreeNode(new WidgetWrapper(container));

		if (container.widgets() == null)
		{
			return root;
		}
		
		for (Widget<?> widget : container.widgets()) 
		{
			if (widget instanceof Container<?>) 
			{
				root.add(buildTreeModel((Container<?>) widget));
			} else 
			{
				root.add(new DefaultMutableTreeNode(new WidgetWrapper(widget)));
			}
		}

		return root;
	}

	/**
	 * Sorts the tree nodes.
	 * 
	 * @param   node
	 *          The tree node where to start the sort.
	 *          
	 * @return  The root node of the sorted tree.
	 */
	private DefaultMutableTreeNode sortNodes(DefaultMutableTreeNode node) 
	{
		for (int i = 0; i < node.getChildCount() - 1; i++) 
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) node.getChildAt(i);
			String nt = child.getUserObject().toString();

			for (int j = i + 1; j <= node.getChildCount() - 1; j++) 
			{
				DefaultMutableTreeNode prevNode = (DefaultMutableTreeNode) node.getChildAt(j);
				String np = prevNode.getUserObject().toString();

				if (nt.compareToIgnoreCase(np) > 0) 
				{
					node.insert(child, j);
					node.insert(prevNode, i);
				}
			}
			if (child.getChildCount() > 0) 
			{
				sortNodes(child);
			}
		}

		return node;
	}

   /**
    * Returns the list of expanded widgets in the tree.
    * 
    * @return  List of widgets.
    */
	private List<Widget<?>> getExpandedWidgets()
   {
      List<Widget<?>> expandedWidgets = new ArrayList<>();
      Enumeration<TreePath> expandedPaths = 
            getExpandedDescendants(new TreePath(getModel().getRoot()));
      
      if (expandedPaths == null)
      {
         return expandedWidgets;
      }

      // iterate per each expanded path
      while (expandedPaths.hasMoreElements())
      {
         // get node for path
         Object node = expandedPaths.nextElement().getLastPathComponent();

         // store widget from node
         if (node instanceof DefaultMutableTreeNode
               && ((DefaultMutableTreeNode) node).getUserObject() instanceof WidgetWrapper)
         {
            expandedWidgets.add(((WidgetWrapper) ((DefaultMutableTreeNode) node).getUserObject())
                  .getWidget());
         }
      }

      return expandedWidgets;
   }

   /**
    * Expand the supplied tree paths.
    * 
    * @param   paths
    *          List of tree paths.
    */
	private void expandRows(List<TreePath> paths)
   {
      for (TreePath path : paths)
      {
         expandPath(path);
      }
   }

	/**
    * Finds tree paths for the supplied list of widgets.
    * 
    * @param   subjects
    *          List of {@link Widget} references.
    *          
    * @return  List of {@link TreePath} references.
    */
   private List<TreePath> findNodes(List<Widget<?>> subjects)
   {
      List<TreePath> output = new ArrayList<>();
      findNodes(getModel(), getModel().getRoot(), subjects, output);
      return output;
   }

   /**
    * Recursively finds tree paths for the supplied list of widgets.
    * 
    * @param   model
    *          {@link TreeModel} reference.
    * @param   node
    *          The root node.
    * @param   subjects
    *          List of {@link Widget} references.
    * @param   output
    *          A valid instance of List of {@link TreePath} references where to store the result. 
    */
   private void findNodes(TreeModel model, Object node, List<Widget<?>> subjects, 
                          List<TreePath> output)
   {
      if (node instanceof DefaultMutableTreeNode
            && ((DefaultMutableTreeNode) node).getUserObject() instanceof WidgetWrapper)
      {
         Widget<?> widget = ((WidgetWrapper) ((DefaultMutableTreeNode) node).getUserObject())
               .getWidget();
         
         for (Widget<?> subject : subjects)
         {
            if (widget == subject)
            {
               TreeNode[] result = ((DefaultMutableTreeNode) node).getPath();

               if (result != null)
               {
                  output.add(new TreePath(result));
               }

               if (subjects.size() == output.size())
               {
                  return;
               }
            }
         }
      }

      for (int i = 0; i < model.getChildCount(node); i++)
      {
         findNodes(model, model.getChild(node, i), subjects, output);

         if (subjects.size() == output.size())
         {
            return;
         }
      }
   }
   
   /**
    * Wrapper class to help store {@link Widget} references 
    * in the tree component.  
    */
   public class WidgetWrapper 
   {
      /**
       * The widget reference.
       */
      private Widget<?> widget;

      /**
       * Constructor.
       * 
       * @param   widget
       *          A valid {@link Widget} reference.
       */
      public WidgetWrapper(Widget<?> widget) 
      {
         this.widget = widget;
      }

      /**
       * Returns the wrapped widget reference.
       * 
       * @return  {@link Widget} reference.
       */
      public Widget<?> getWidget() 
      {
         return widget;
      }
      
      /**
       * Builds map of properties to display in the property table.
       * 
       * @return  map of string-object entries.
       */
      public Map<String, Object> getProperties()
      {
         Map<String, Object> result = new HashMap<>();
         
         String[] propNames =
             { "ancestor", "ancestorId", "bounds", "dimension", "hasFocus", "height", "hidden",
               "highlighted", "insets", "isDisplayed", "isEnabled", "isFocusable", "isHeader",
               "isInheritHeight", "isInheritWidth", "isTotallyObscured", "isVisible", "location",
               "minimumSize", "name", "parent", "physicalBounds", "physicalDimension",
               "physicalLocation", "screenLocation", "width" };         
         
         for (String name : propNames) 
         {
            Object value = null;
            boolean found = false;

            try 
            {
               try 
               {
                  // load field
                  Field field = widget.getClass().getField(name);
                  field.setAccessible(true);;
                  value = formatValue(field.get(widget));
                  found = true;
               } 
               catch (NoSuchFieldException ex) 
               {
                  // field doesn't exist
               }
   
               try 
               {
                  // load method
                  Method method = widget.getClass().getMethod(name);
                  method.setAccessible(true);
                  value = formatValue(method.invoke(widget));
                  found = true;
               } 
               catch (NoSuchMethodException ex) 
               {
                  // method doesn't exist
               }
               
               if (!found) 
               {
                  // neither field nor method found
                  value = "<reflection error, method or field not found>";
               }
            }
            catch (InvocationTargetException ex)
            {
               value = "<" + ex.getCause().toString() + ">";
            }
            catch (Exception ex)
            {
               value = "<" + ex.toString() + ">";
            }

            // add object, toString() is called automatically
            result.put(name, value);
         }
         
         // fields of the associated config
         WidgetConfig config = widget.config();
         if (config != null)
         {
            Class<?> cfgClass = config.getClass();
            while(!cfgClass.equals(Object.class))
            {
               Field[] fields = cfgClass.getDeclaredFields();
               for (Field field : fields)
               {
                  // skip static fields
                  if (java.lang.reflect.Modifier.isStatic(field.getModifiers()))
                  {
                     continue;
                  }
                  
                  String fieldName = field.getName();
                  
                  // skip aspectj fields
                  if (fieldName.startsWith("ajc$"))
                  {
                     continue;
                  }
                  
                  field.setAccessible(true);
                  Object fieldValue = null;
                  try
                  {
                     fieldValue = field.get(config);
                  }
                  catch (Exception e)
                  {
                     fieldValue = "<" + e.toString() + ">";
                  }
                  result.put("config." + fieldName, fieldValue);
               }
               cfgClass = cfgClass.getSuperclass();
            }
         }
         
         // object identity hash-code property
         String hashcode = Integer.toHexString(System.identityHashCode(widget));
         result.put("hashcode", hashcode);
         
         return result;
      }
      
      /**
       * This override is called by the tree component to show the node label.
       * 
       * @return  String representing this widget in the tree component.
       */
      @Override
      public String toString() 
      {
         return formatWidget(widget);
      }
      
      /**
       * Formats the supplied reference into a string.
       * 
       * @param   obj
       *          A reference, may be <code>null</code>.
       * @return  The formatted value.
       */
      private String formatValue(Object obj)
      {
         if (obj instanceof Widget<?>)
         {
            return formatWidget((Widget<?>) obj);
         }
         
         if (obj instanceof Point ||
             obj instanceof NativePoint ||
             obj instanceof Dimension ||
             obj instanceof NativeDimension ||
             obj instanceof Insets ||
             obj instanceof Rectangle ||
             obj instanceof NativeRectangle)
         {
            return ToStringBuilder.reflectionToString(obj, ToStringStyle.SHORT_PREFIX_STYLE);
         }
         
         return Objects.toString(obj);
      }

      /**
       * Formats the supplied reference into a string.
       * 
       * @param   widget
       *          A valid {@link Widget} reference.
       *          
       * @return  The formatted string.
       */
      private String formatWidget(Widget<?> widget)
      {
         String className = showFullClassName ? widget.getClass().getName() : widget.getClass()
               .getSimpleName();

         String id = widget.config() == null || widget.config().id == null ? "null" : widget
               .config().id.toString();

         return String.format("%s<%s>", className, id);
      }
   }
}
